#######################################
# CONSTANTS
#######################################


TT_INT = 'INT'
TT_FLOAT = 'FLOAT'
TT_STRING = 'STRING'
TT_NUMERAL = 'NUMERAL'
TT_IDENTIFIER = 'IDENTIFIER'
TT_DOT = 'DOT'
TT_KEYWORD = 'KEYWORD'
TT_PLUS = 'PLUS'
TT_MINUS = 'MINUS'
TT_MUL = 'MUL'
TT_DIV = 'DIV'
TT_POW = 'POW'
TT_EQ = 'EQ'
TT_PLUS_EQ = 'PLUS_EQ'
TT_MIN_EQ = 'MIN_EQ'
TT_MUL_EQ = "MUL_EQ"
TT_DIV_EQ = "DIV_EQ"
TT_LPAREN = 'LPAREN'
TT_RPAREN = 'RPAREN'
TT_LSQUARE = 'LSQUARE'
TT_RSQUARE = 'RSQUARE'
TT_LBRACE = 'LBRACE'
TT_RBRACE = 'RBRACE'
TT_EE = 'EE'
TT_NE = 'NE'
TT_LT = 'LT'
TT_GT = 'GT'
TT_LTE = 'LTE'
TT_GTE = 'GTE'
TT_COLON = 'COLON'
TT_COMMA = 'COMMA'
TT_ARROW = 'ARROW'
TT_TAB = 'TAB'
TT_NEWLINE = 'NEWLINE'
TT_EOF = 'EOF'

KEYWORDS = [
    'Verus',
    'Falsus',
    'et',
    'aut',
    'non',
    'si',
    'alioquinsi',
    'alioquin',
    'tempta',
    'praeter',
    'tam',
    'pro',
    'ad',
    'gradus',
    'dum',
    'opus',
    'redi',
    'continua',
    'confringe',
]
